// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:equatable/equatable.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:wyatt_authentication_bloc/src/core/enum/auth_cubit_status.dart';
import 'package:wyatt_authentication_bloc/src/core/exceptions/exceptions.dart';
import 'package:wyatt_authentication_bloc/src/domain/repositories/authentication_repository.dart';
import 'package:wyatt_form_bloc/wyatt_form_bloc.dart';

part 'sign_in_state.dart';

class SignInCubit extends Cubit<SignInState> {
  final AuthenticationRepository _authenticationRepository;

  final FormValidator _validationStrategy;

  SignInCubit({
    required AuthenticationRepository authenticationRepository,
    FormValidator validationStrategy = const EveryInputValidator(),
  })  : _authenticationRepository = authenticationRepository,
        _validationStrategy = validationStrategy,
        super(const SignInState());

  void emailChanged(String value) {
    final Email email = Email.dirty(value);
    emit(
      state.copyWith(
        email: email,
        status: _validationStrategy.rawValidate([email, state.password]),
      ),
    );
  }

  void passwordChanged(String value) {
    final Password password = Password.dirty(value);
    emit(
      state.copyWith(
        password: password,
        status: _validationStrategy.rawValidate([state.email, password]),
      ),
    );
  }

  Future<void> signInAnonymously() async {
    if (state.status.isSubmissionInProgress) {
      return;
    }

    emit(state.copyWith(status: FormStatus.submissionInProgress));
    try {
      await _authenticationRepository.signInAnonymously();
      _authenticationRepository.changeCubitStatus(AuthCubitStatus.started);
      emit(state.copyWith(status: FormStatus.submissionSuccess));
    } on SignInAnonymouslyFailureInterface catch (e) {
      emit(
        state.copyWith(
          errorMessage: e.message,
          status: FormStatus.submissionFailure,
        ),
      );
    } catch (_) {
      emit(state.copyWith(status: FormStatus.submissionFailure));
    }
  }

  Future<void> signInWithGoogle() async {
    if (state.status.isSubmissionInProgress) {
      return;
    }

    emit(state.copyWith(status: FormStatus.submissionInProgress));
    try {
      await _authenticationRepository.signInWithGoogle();
      _authenticationRepository.changeCubitStatus(AuthCubitStatus.started);
      emit(state.copyWith(status: FormStatus.submissionSuccess));
    } on SignInWithGoogleFailureInterface catch (e) {
      emit(
        state.copyWith(
          errorMessage: e.message,
          status: FormStatus.submissionFailure,
        ),
      );
    } catch (_) {
      emit(state.copyWith(status: FormStatus.submissionFailure));
    }
  }

  Future<void> signInWithEmailAndPassword() async {
    if (!state.status.isValidated) {
      return;
    }
    emit(state.copyWith(status: FormStatus.submissionInProgress));
    try {
      await _authenticationRepository.signInWithEmailAndPassword(
        email: state.email.value,
        password: state.password.value,
      );
      _authenticationRepository.changeCubitStatus(AuthCubitStatus.started);
      emit(state.copyWith(status: FormStatus.submissionSuccess));
    } on SignInWithEmailAndPasswordFailureInterface catch (e) {
      emit(
        state.copyWith(
          errorMessage: e.message,
          status: FormStatus.submissionFailure,
        ),
      );
    } catch (_) {
      emit(state.copyWith(status: FormStatus.submissionFailure));
    }
  }
}
