// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:wyatt_authentication_bloc/src/core/enum/auth_cubit_status.dart';
import 'package:wyatt_authentication_bloc/src/core/exceptions/exceptions.dart';
import 'package:wyatt_authentication_bloc/src/domain/repositories/authentication_repository.dart';
import 'package:wyatt_form_bloc/wyatt_form_bloc.dart';

part 'sign_up_state.dart';

class SignUpCubit extends Cubit<SignUpState> {
  final AuthenticationRepository _authenticationRepository;

  final Future<void> Function(SignUpState state, String? uid)? _onSignUpSuccess;
  final FormValidator _validationStrategy;

  SignUpCubit({
    required AuthenticationRepository authenticationRepository,
    required FormData formData,
    FormValidator validationStrategy = const EveryInputValidator(),
    Future<void> Function(SignUpState state, String? uid)? onSignUpSuccess,
  })  : _authenticationRepository = authenticationRepository,
        _onSignUpSuccess = onSignUpSuccess,
        _validationStrategy = validationStrategy,
        super(SignUpState(data: formData));

  void emailChanged(String value) {
    final Email email = Email.dirty(value);

    final List<FormInputValidator<dynamic, ValidationError>> toValidate = [
      email,
      state.password,
      ...state.data.validators<dynamic, ValidationError>(),
    ];

    emit(
      state.copyWith(
        email: email,
        status: _validationStrategy.rawValidate(toValidate),
      ),
    );
  }

  void passwordChanged(String value) {
    final Password password = Password.dirty(value);

    final List<FormInputValidator<dynamic, ValidationError>> toValidate = [
      state.email,
      password,
      ...state.data.validators<dynamic, ValidationError>(),
    ];

    emit(
      state.copyWith(
        password: password,
        status: _validationStrategy.rawValidate(toValidate),
      ),
    );
  }

  // Take from wyatt_form_bloc/wyatt_form_bloc.dart
  void dataChanged<T>(
    String field,
    FormInputValidator<T, ValidationError> dirtyValue,
  ) {
    final form = state.data.clone();

    if (form.contains(field)) {
      form.updateValidator(field, dirtyValue);
    } else {
      throw Exception('Form field $field not found');
    }

    emit(
      state.copyWith(
        data: form,
        status: _validationStrategy.rawValidate(
          [
            state.email,
            state.password,
            ...state.data.validators<dynamic, ValidationError>(),
          ],
        ),
      ),
    );
  }

  // Take from wyatt_form_bloc/wyatt_form_bloc.dart
  void updateFormData(
    FormData data, {
    SetOperation operation = SetOperation.replace,
  }) {
    FormData form = data;

    switch (operation) {
      case SetOperation.replace:
        form = data;
        break;
      case SetOperation.difference:
        form = state.data.difference(data);
        break;
      case SetOperation.intersection:
        form = state.data.intersection(data);
        break;
      case SetOperation.union:
        form = state.data.union(data);
        break;
    }

    emit(
      state.copyWith(
        data: form,
        status: _validationStrategy.rawValidate(
          [
            state.email,
            state.password,
            ...state.data.validators<dynamic, ValidationError>(),
          ],
        ),
      ),
    );
  }

  Future<void> signUpFormSubmitted() async {
    if (!state.status.isValidated) {
      return;
    }
    emit(state.copyWith(status: FormStatus.submissionInProgress));
    try {
      final uid = await _authenticationRepository.signUp(
        email: state.email.value,
        password: state.password.value,
      );
      await _onSignUpSuccess?.call(state, uid);
      _authenticationRepository.changeCubitStatus(AuthCubitStatus.started);
      emit(state.copyWith(status: FormStatus.submissionSuccess));
    } on SignUpWithEmailAndPasswordFailureInterface catch (e) {
      emit(
        state.copyWith(
          errorMessage: e.message,
          status: FormStatus.submissionFailure,
        ),
      );
    } catch (e) {
      debugPrint(e.toString());
      emit(state.copyWith(status: FormStatus.submissionFailure));
    }
  }
}
