// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:wyatt_crud_bloc/src/models/model.dart';
import 'package:wyatt_crud_bloc/src/models/queries/queries_firestore.dart';
import 'package:wyatt_crud_bloc/src/models/queries/queries_interface.dart';
import 'package:wyatt_crud_bloc/src/repositories/crud_repository_interface.dart';

class CrudRepositoryFirestore<T> implements CrudRepositoryInterface<T> {
  final FirebaseFirestore _firestore;
  final Model<DocumentSnapshot, T> _parser;

  late CollectionReference _collectionReference;

  CrudRepositoryFirestore(
    String collection,
    Model<DocumentSnapshot, T> parser, {
    FirebaseFirestore? firestore,
  })  : _firestore = firestore ?? FirebaseFirestore.instance,
        _parser = parser {
    _collectionReference = _firestore.collection(collection);
  }

  @override
  Future<void> create(Model object, {String? id}) {
    if (id != null) {
      return _collectionReference.doc(id).set(object.toMap());
    } else {
      return _collectionReference.add(object.toMap());
    }
  }

  @override
  Future<void> delete(String id) {
    return _collectionReference.doc(id).delete();
  }

  @override
  Future<void> deleteAll() async {
    final _batch = _firestore.batch();
    final QuerySnapshot snapshots = await _collectionReference.get();
    for (final DocumentSnapshot snapshot in snapshots.docs) {
      _batch.delete(snapshot.reference);
    }
    return _batch.commit();
  }

  @override
  Future<T?> get(String id) async {
    final DocumentSnapshot snapshot = await _collectionReference.doc(id).get();
    return _parser.from(snapshot);
  }

  @override
  Future<List<T?>> getAll() async {
    final QuerySnapshot snapshots = await _collectionReference.get();
    return snapshots.docs.map(_parser.from).toList();
  }

  @override
  Future<List<T?>> query(List<QueryInterface> conditions) async {
    Query query = _collectionReference;
    for (final condition in conditions) {
      query = QueryParserFirestore().parser(condition, query);
    }
    final QuerySnapshot snapshots = await query.get();
    return snapshots.docs.map(_parser.from).toList();
  }

  @override
  Stream<List<T?>> stream({
    String? id,
    List<QueryInterface>? conditions,
    bool includeMetadataChanges = false,
  }) {
    if (id != null) {
      return _collectionReference
          .doc(id)
          .snapshots(
            includeMetadataChanges: includeMetadataChanges,
          )
          .map<List<T?>>(
            (DocumentSnapshot snapshot) => [_parser.from(snapshot)],
          );
    } else {
      if (conditions != null) {
        Query query = _collectionReference;
        for (final condition in conditions) {
          query = QueryParserFirestore().parser(condition, query);
        }
        return query
            .snapshots(
          includeMetadataChanges: includeMetadataChanges,
        )
            .map((querySnapshot) {
          return querySnapshot.docs.map(_parser.from).toList();
        });
      } else {
        return _collectionReference
            .snapshots(
          includeMetadataChanges: includeMetadataChanges,
        )
            .map((querySnapshot) {
          return querySnapshot.docs.map(_parser.from).toList();
        });
      }
    }
  }

  @override
  Future<void> update(String id, {Model? object, Map<String, Object?>? raw}) {
    if (object != null) {
      return _collectionReference.doc(id).update(object.toMap());
    } else {
      if (raw != null) {
        return _collectionReference.doc(id).update(raw);
      } else {
        throw Exception('You must provide an object or a raw map');
      }
    }
  }

  @override
  Future<void> updateAll(Map<String, Object?> raw) async {
    final _batch = _firestore.batch();
    final QuerySnapshot snapshots = await _collectionReference.get();
    for (final DocumentSnapshot snapshot in snapshots.docs) {
      _batch.update(snapshot.reference, raw);
    }
    return _batch.commit();
  }
}
